#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <a.out.h>
#include <sys/param.h>
#include <sys/exec.h>
#include "ld.h"

char *input_name;

#define TEXT_PTR(x)	(text_base + (x))

char *type_letters = "uUaAtTdDbB        cC          fF";

main(ac, av)
int ac;
char **av;
{
    int desc, nsymbols, string_size;
    int	n, i, has_nz = 0;
    struct nlist *np;
    struct nzlist *nzp;
    char *strings, *contents;
    struct link_object *lop;
    long next, txtaddr;
    char *text_base, *name;
    struct exec hdr;
    struct link_dynamic dyn;
    struct link_dynamic_2 dyn2;
    int nreloc;
    struct relocation_info *reloc;

    if (ac != 2) {
	fprintf(stderr, "Usage: %s exec-file\n", av[0]);
	exit(1);
    }
    input_name = av[1];
    if ((desc = open(input_name, O_RDONLY, 0)) < 0) {
	perror(input_name);
	exit(1);
    }

    if (read(desc, &hdr, sizeof(hdr)) != sizeof(hdr))
	die("failure reading header");
    md_swapin_exec_hdr(&hdr);

    if (N_BADMAG(hdr))
	die("malformed");
    if (!N_IS_DYNAMIC(hdr))
	die("not dynamic");

    n = hdr.a_text + hdr.a_data;
    contents = alloca(n);
    if (read(desc, contents, n) != n)
	die("read error");
    txtaddr = (hdr.a_entry < __LDPGSZ)? 0: N_TXTADDR(hdr);
    text_base = contents - sizeof(hdr) + N_TXTOFF(hdr) - txtaddr;

    /* Read DYNAMIC structure (first in data segment) */
    dyn = *(struct link_dynamic *)TEXT_PTR(txtaddr + hdr.a_text);
    md_swapin_link_dynamic(&dyn);

    printf("ld_dynamic: version=%d, ldd=%x, ld_2=%x, ld_entry=%x\n",
	   dyn.ld_version, dyn.ldd, dyn.ld_un.ld_2, dyn.ld_entry);

    /* Check version */
    switch (dyn.ld_version) {
    default:
	die("unsupported _DYNAMIC version");
	break;
    case LD_VERSION_SUN:
	break;
    case LD_VERSION_BSD:
	has_nz = 1;
	break;
    }

    /* Read link_dynamic_2 struct (from data segment) */
    dyn2 = *(struct link_dynamic_2 *)TEXT_PTR((long)dyn.ld_un.ld_2);
    md_swapin_link_dynamic_2(&dyn2);

    printf("dynamic_2: ld_loaded=%x ld_need=%x ld_rules=%x ld_got=%x\n",
	   dyn2.ld_loaded, dyn2.ld_need, dyn2.ld_rules, dyn2.ld_got);
    printf("	ld_plt=%x ld_rel=%x ld_hash=%x ld_symbols=%x\n",
	   dyn2.ld_plt, dyn2.ld_rel, dyn2.ld_hash, dyn2.ld_symbols);
    printf("	ld_stab_hash=%x ld_buckets=%x ld_strings=%x ld_str_sz=%x\n",
	   dyn2.ld_stab_hash, dyn2.ld_buckets, dyn2.ld_strings, dyn2.ld_str_sz);
    printf("	ld_text_sz=%x ld_plt_sz=%x\n", dyn2.ld_text_sz, dyn2.ld_plt_sz);

    printf("\nObjects required:\n");
    next = dyn2.ld_need;
    while (next != 0) {
	lop = (struct link_object *)TEXT_PTR(next);
	md_swapin_link_object(lop, 1);
	name = TEXT_PTR(lop->lo_name);
	if (lop->lo_library)
	    printf("-l%s.%d.%d\n", name, lop->lo_major, lop->lo_minor);
	else
	    printf("%s\n", name);
	next = lop->lo_next;
    }

    /* Read symbols (text segment) */
    n = dyn2.ld_strings - dyn2.ld_symbols;
    nsymbols = n / (has_nz ? sizeof(struct nzlist) : sizeof(struct nlist));
    nzp = (struct nzlist *)TEXT_PTR(dyn2.ld_symbols);
    np = (struct nlist *) nzp;
    if (has_nz)
	md_swapin_zsymbols(nzp, nsymbols);
    else
	md_swapin_symbols(np, nsymbols);

    /* Read strings (text segment) */
    n = string_size = dyn2.ld_str_sz;
    strings = (char *)TEXT_PTR(dyn2.ld_strings);

    printf("\nSymbols:\ntype\tother\tdesc\tvalue\tsize\tname\n");
    for (i = 0; i < nsymbols; ++i) {
	n = has_nz? nzp[i].nz_type: np[i].n_type;
	if (n < 0x20 && type_letters[n] != ' ')
	    printf("%c", type_letters[n]);
	else
	    printf("%.2x", n);
	if (has_nz)
	    printf("\t%x\t%x\t%x\t%x\t%s\n",
		   nzp[i].nz_other, nzp[i].nz_desc, nzp[i].nz_value,
		   nzp[i].nz_size, strings + nzp[i].nz_strx);
	else
	    printf("\t%x\t%x\t%x\t\t%s\n",
		   np[i].n_other, np[i].n_desc, np[i].n_value,
		   strings + np[i].n_un.n_strx);
    }

    nreloc = (dyn2.ld_hash - dyn2.ld_rel) / sizeof (struct relocation_info);
    printf ("\n%d Relocs:\n", nreloc);
    n = 0;
    for (reloc = (struct relocation_info *) TEXT_PTR(dyn2.ld_rel);
	 nreloc;
	 nreloc--, reloc++) 
      printf ("%8d  0x%x\n", n++, reloc->r_address);
    printf ("\n");
}

die(str)
    char *str;
{
    fprintf(stderr, "%s: %s\n", input_name, str);
    exit(1);
}
